/**
 * [BSD Open Source License]
 *
 * Copyright (c) 2006, Mattias Jiderhamn, Expert Systems
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or othermaterials provided with the distribution.
 *
 * Neither the name of Expert Systems nor the names of its contributors may
 * be usedto endorse or promote products derived from this software without
 * specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package se.expertsystem.intellij.aop.aspectj;

import com.intellij.openapi.diagnostic.Logger;
import com.intellij.openapi.module.Module;
import com.intellij.openapi.module.ModuleManager;
import com.intellij.openapi.progress.ProgressIndicator;
import com.intellij.openapi.progress.ProgressManager;
import com.intellij.openapi.project.Project;
import com.intellij.openapi.vfs.JarFileSystem;
import com.intellij.openapi.vfs.VirtualFileManager;
import org.aspectj.apache.bcel.classfile.ClassParser;
import org.aspectj.apache.bcel.classfile.JavaClass;
import org.aspectj.apache.bcel.classfile.ClassFormatException;
import org.aspectj.util.FileUtil;
import org.aspectj.weaver.ReferenceType;
import org.aspectj.weaver.UnresolvedType;
import org.aspectj.weaver.bcel.BcelObjectType;
import org.aspectj.weaver.bcel.BcelWorld;
import se.expertsystem.intellij.PluginUtils;

import java.io.*;
import java.net.URL;
import java.util.*;
import java.util.zip.ZipEntry;
import java.util.zip.ZipInputStream;

/**
 * This class finds and reads aspect classes.
 * @author Mattias Jiderhamn
 */
public abstract class AspectFinder {

  private static final Logger LOG = Logger.getInstance(AspectFinder.class.getName());

  /**
   * Find location of all (compiled) aspects in the project.
   * Searches output paths and libraries/classpath.
   * @return SortedMap with aspect class name as key and URL (String) to class file as value.
   */
  static SortedMap/*<String,String>*/ findAspectLocations(Project project) {
    SortedMap aspectLocations = new TreeMap();
    Map foundAspects = findAspects(project);
    // Add as known aspects
    for (Iterator i = foundAspects.keySet().iterator(); i.hasNext();) {
      ReferenceType aspectX = (ReferenceType) i.next();
      // LOG.info("Aspect: " + aspectX.getName() + " in file " + fileName);
      aspectLocations.put(aspectX.getName(), foundAspects.get(aspectX));
    }
    return aspectLocations;
  }

  /**
   * Find all (compiled) aspects in the project.
   * Searches output paths and libraries/classpath.
   * @return Map with aspect as key and URL (String) to class file as value.
   */
  private static Map/*<String,ReferenceType>*/ findAspects(Project project) {
    Module[] modules = ModuleManager.getInstance(project).getModules();
    try {
      final BcelWorld world = new BcelWorld(); // Dummy World
      world.setBehaveInJava5Way(PluginUtils.isJava5(project));
      return findAspects(modules, world);
    }
    catch(IOException ioex) {
      LOG.error(ioex);
      return new HashMap();
    }
  }
  
  /**
   * Find all (compiled) aspects in the modules.
   * Searches output paths and libraries/classpath.
   * @return Map with aspect as key and URL (String) to class file as value.
   */
  private static Map/*<String,ReferenceType>*/ findAspects(Module[] modules, final BcelWorld world) throws IOException {
    Map allAspects = new HashMap();
    ProgressIndicator progressIndicator = ProgressManager.getInstance().getProgressIndicator();

    // Collect directories/JARs to look for aspects in
    Set/*<File>*/ filesToProcess = new com.intellij.util.containers.HashSet();
    for(int i = 0; i < modules.length; i++) {
      Module module = modules[i];
      List classpath = PluginUtils.getModuleClasspath(module, false); // (Exclude JDK)

      for(Iterator cpIter = classpath.iterator(); cpIter.hasNext(); ) {
        File file = (File) cpIter.next();
        filesToProcess.add(file);
        world.addPath(file.getAbsolutePath()); // Add to world for automatic resolving
      }
    }

    // Look for aspects in directories/JARs
    final int noOfFiles = filesToProcess.size();
    int count = 0;
    progressIndicator.setIndeterminate(false);
    progressIndicator.setText("Searching for aspects");
    for(Iterator iter = filesToProcess.iterator(); iter.hasNext(); ) {
      if(progressIndicator.isCanceled())
        return allAspects; // Return what we have found so far
      PluginUtils.setProgress(progressIndicator, noOfFiles, ++count);
      File file = (File) iter.next();
      progressIndicator.setText2("Looking for aspects in '" + file.getName() + "'");
      LOG.debug("Looking for aspects in " + file);
      allAspects.putAll(findAspects(file, world));
    }
    progressIndicator.setIndeterminate(true);
    return allAspects;
  }

  /**
   * Find all (compiled) aspects in the provided file.
   * Searches output paths and libraries/classpath.
   * @param  file May be a directory or a JAR file containing .class files.
   * @return Map with aspect as key and URL (String) to class file as value.
   */
  private static Map/*<String,ReferenceType>*/ findAspects(File file, final BcelWorld world) throws IOException {
    // The operation below corresponds to
    //   weaver.addLibraryJarFile(file);
    // except that we don't keep all classes in memory, to avoid OutOfMemoryError
    Map foundAspects = new HashMap();
    if(! file.exists()) {
      // Do nothing; no aspects
    }
    else if (file.isDirectory()) {
      // Find all class files in directory
      File[] classFiles = FileUtil.listFiles(file,new FileFilter(){
        public boolean accept(File pathname) {
          return pathname.getName().endsWith(".class");
        }
      });

      // Find aspects among classes
      for (int i = 0; i < classFiles.length; i++) {
        final File classFile = classFiles[i];
        try {
          LOG.debug(classFile.getAbsolutePath()); // TODO Remove
          ReferenceType type = readAspect(classFile, world);
          if (type.isAspect() && ! type.isAbstract())
            foundAspects.put(type, PluginUtils.toUrlString(classFile));
        } catch(ClassFormatException cfex) {
          LOG.info(cfex);
        }
      }
    }
    else { // JAR file
      ZipInputStream zis = null; 
      try {
        zis = new ZipInputStream(new FileInputStream(file));
        while (true) {
          ZipEntry zipEntry = zis.getNextEntry();
          if (zipEntry == null) break;

          if (zipEntry.isDirectory() || !zipEntry.getName().endsWith(".class") ||
              zipEntry.getName().equals("org/aspectj/lib/pointcuts/Pointcuts.class") || // Some kind of error in there!
              zipEntry.getName().equals("org/aspectj/lib/tracing/TraceJoinPointsBase.class")) { // Some kind of error in there!
            continue;
          }

          LOG.debug(file.getAbsolutePath() + '/' + zipEntry.getName()); // TODO Remove
          ReferenceType type = resolveType(FileUtil.readAsByteArray(zis), zipEntry.getName(), world);
          zis.closeEntry();

          if (type.isAspect() && ! type.isAbstract()) {
            // Construct JAR url
            String fileUrl = PluginUtils.toUrlString(file);
            // (Same as JarFileSystem.PROTOCOL + "://" + PluginUtils.toUrlString(file))
            String jarUrl = VirtualFileManager.constructUrl(JarFileSystem.PROTOCOL, fileUrl) +
                JarFileSystem.JAR_SEPARATOR + zipEntry.getName();
            foundAspects.put(type, jarUrl);
          }
        }
      }
      catch(IOException ioex) { // Probably ZipException
        LOG.error(ioex);
        // Doesn't seem to work with progress indication
        /*
        Messages.showWarningDialog(
          "There was an error parsing\n" + file.getAbsolutePath(),
          "ZIP error");
        */
      }
      catch(ClassFormatException cfex) {
        LOG.info(cfex);
      }
      finally {
        if(zis != null)
          zis.close();
      }
    }

    return foundAspects;
  }


  /**
   * Read aspect from specified url.
   * @param url May be file URL to .class file, or jar URL to class file inside JAR file.
   */
  static ReferenceType readAspect(final String url, final BcelWorld world) throws IOException {
    LOG.debug("protocol of " + url + ": " + VirtualFileManager.extractProtocol(url) );
    if(JarFileSystem.PROTOCOL.equals(VirtualFileManager.extractProtocol(url))) { // JAR url
      String path = url.substring((JarFileSystem.PROTOCOL + "://").length());
      final int delimIndex = path.indexOf(JarFileSystem.JAR_SEPARATOR);
      String jarPath = path.substring(0, delimIndex);
      String entryName = path.substring(delimIndex + JarFileSystem.JAR_SEPARATOR.length());
      ZipInputStream zis = null;
      try {
        zis = new ZipInputStream(new FileInputStream(PluginUtils.toFile(jarPath)));
        while (true) {
          ZipEntry zipEntry = zis.getNextEntry();
          if (zipEntry == null)
            return null;
          else if(zipEntry.getName().equals(entryName)) {
            ReferenceType type = resolveType(FileUtil.readAsByteArray(zis), zipEntry.getName(), world);
            zis.closeEntry();
            return type;
          }
        }
      }
      catch(IOException ioex) {
        LOG.error(ioex);
        // Doesn't seem to work with progress indication
        /*
        Messages.showWarningDialog(
          "There was an error parsing\n" + jarPath,
          "ZIP error");
        */
      }
      catch (ClassFormatException cfex){
        LOG.info(cfex);
      }
      finally {
        if(zis != null)
          zis.close();
      }
      return null;
    }
    else {
      final File classFile = PluginUtils.toFile(new URL(url));
      if (classFile.exists()) {
        try {
          return readAspect(classFile, world);
        } catch(ClassFormatException cfex) {
          LOG.info(cfex);
          return null;
        }
      } else {
        return null;
      }
    }
  }

  /** Read aspect from specified file */
  private static ReferenceType readAspect(final File classFile, final BcelWorld world) throws IOException, ClassFormatException {
    byte[] bytes = FileUtil.readAsByteArray(classFile);
    String name = classFile.getAbsolutePath();
    ReferenceType type = resolveType(bytes, name, world);
    return type;
  }

  /** Convert byte array to aspect representation  */
  private static ReferenceType resolveType(byte[] bytes, String name, BcelWorld world) throws IOException, ClassFormatException {
    ClassParser parser = new ClassParser(new ByteArrayInputStream(bytes),name);
    JavaClass jc = parser.parse();
    return getResolvedType(jc, world);
  }

  /**
   * Resolve type (aspect representation) from parsed Java class. Used to avoid calling
   *   <code>BcelWorld.addSourceObjectType(jc).getResolvedTypeX();</code>
   *  since it will use a lot of memory for large project.
   */
  static ReferenceType getResolvedType(JavaClass jc, BcelWorld world) {
    ReferenceType nameTypeX;
    String signature = UnresolvedType.forName(jc.getClassName()).getSignature();
    if (jc.isGeneric() && world.isInJava5Mode()) {
      nameTypeX =  ReferenceType.fromTypeX(UnresolvedType.forRawTypeName(jc.getClassName()),world);
      BcelObjectType ret = world.buildBcelDelegate(nameTypeX, jc, false, true);
      ReferenceType genericRefType = new ReferenceType(
          UnresolvedType.forGenericTypeSignature(signature,ret.getDeclaredGenericSignature()),world);
      nameTypeX.setDelegate(ret);
      genericRefType.setDelegate(ret);
      nameTypeX.setGenericType(genericRefType);
    } else {
      nameTypeX = new ReferenceType(signature, world);
      world.buildBcelDelegate(nameTypeX, jc, false, true); // Will set delegate
    }
    return nameTypeX;
  }
}
