/**
 * [BSD Open Source License]
 *
 * Copyright (c) 2006, Mattias Jiderhamn, Expert Systems
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or othermaterials provided with the distribution.
 *
 * Neither the name of Expert Systems nor the names of its contributors may
 * be usedto endorse or promote products derived from this software without
 * specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package se.expertsystem.intellij.aop.aspectj;

import com.intellij.openapi.diagnostic.Logger;
import com.intellij.openapi.project.Project;
import com.intellij.openapi.ui.Messages;
import com.intellij.util.containers.HashSet;

import javax.swing.*;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import java.awt.*;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.util.Map;
import java.util.Set;
import java.util.SortedMap;
import java.util.TreeMap;

import se.expertsystem.intellij.PluginUtils;

/**
 * This class creates the configuration dialog for the AspectJ weaver component.
 * @author Mattias Jiderhamn
 */
public class AspectJConfiguration extends JComponent {

  private final static Logger LOG = Logger.getInstance(AspectJConfiguration.class.getName());

  /** Is the component enabled? */
  private boolean enabled;

  /** Map from class name to class file url for all known aspects*/
  private final SortedMap/*<String,String>*/ aspectLocations = new TreeMap();

  /** Set of class names for enabled aspects */
  private final Set/*<String>*/ enabledAspects = new HashSet();

  /** Get state of enabled checkbox in configuration dialog */
  public boolean isEnabled() {
    return enabled;
  }

  /** Get locations of all known aspects from configuration dialog */
  public SortedMap/*<String,String>*/ getAspectLocations() {
    return aspectLocations;
  }

  /** Get name of all enabled aspects from configuration dialog*/
  public Set/*<String>*/ getEnabledAspects() {
    return enabledAspects;
  }

  public AspectJConfiguration(final Project project, boolean enabled, Map aspLocs, Set enbAsps) {
    this.enabled = enabled;
    this.aspectLocations.putAll(aspLocs); // Copy locations, to detect changes
    this.enabledAspects.addAll(enbAsps); // Copy names, to detect changes

    setLayout(new BorderLayout());

    // Enable checkbox at the top
    JCheckBox enabledBox = new JCheckBox("Enable weaving");
    enabledBox.setSelected(enabled);
    enabledBox.addChangeListener(new ChangeListener() {
      public void stateChanged(ChangeEvent e) {
        AspectJConfiguration.this.enabled = ((JCheckBox)e.getSource()).isSelected();
      }
    });
    add(enabledBox, BorderLayout.NORTH);

    // List of known aspects to enable/disable in the middle
    JPanel middlePane = new JPanel();
    middlePane.setLayout(new BorderLayout());
    middlePane.setBorder(BorderFactory.createTitledBorder("Aspects to weave"));
    final JList list = new JList(getAspectArray());
    list.setCellRenderer(new CheckboxListCellRenderer());
    list.addMouseListener(new MouseAdapter() {
      public void mouseClicked(MouseEvent e) {
        int index = list.locationToIndex(e.getPoint());
        String className = getClassName(index); // Find class name for clicked item
        if(enabledAspects.contains(className))
          enabledAspects.remove(className);
        else
          enabledAspects.add(className);
        Rectangle rect = list.getCellBounds(index, index);
        list.repaint(rect); // Repaint new state of checkbox
      }
    });
    middlePane.add(new JScrollPane(list), BorderLayout.CENTER); // Make list scrollable
    add(middlePane, BorderLayout.CENTER);

    // Search button at the bottom
    final JButton searchButton = new JButton("Search for aspects...");
    searchButton.addActionListener( new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        Messages.showInfoMessage("Note! Aspects must be compiled before they are found.",
            "Apsects must be compiled");
        // Show progressbar while searching for aspects
        PluginUtils.runProcessWithProgressSynchronously(new Runnable() {
          public void run() {
            SortedMap foundAspects = AspectFinder.findAspectLocations(project);
            Set newAspects = new HashSet(foundAspects.keySet());
            newAspects.removeAll(AspectJConfiguration.this.aspectLocations.keySet()); // Remove all previously known aspects
            AspectJConfiguration.this.aspectLocations.clear();
            AspectJConfiguration.this.aspectLocations.putAll(foundAspects);
            AspectJConfiguration.this.enabledAspects.addAll(newAspects); // Set all new as enabled
            list.setListData(getAspectArray());
            list.repaint();
          }
        }, "Searching for aspects", true, project);
        LOG.debug("Aspects after search: " + AspectJConfiguration.this.aspectLocations);
      }
    });
    add(searchButton, BorderLayout.SOUTH);
  }

  /** Get String array of aspect names, used as list data */
  private String[] getAspectArray() {
    return (String[]) aspectLocations.keySet().toArray(new String[0]);
  }

  /** Find class name given list index */
  private String getClassName(int row) {
    return getAspectArray()[row];
  }

  /* List renderer that shows a checkbox for each known aspect, checked if enabled. */
  private class CheckboxListCellRenderer extends JCheckBox implements ListCellRenderer {
    public Component getListCellRendererComponent(JList list, Object value, final int index, boolean isSelected, boolean cellHasFocus) {
      final String className = value.toString();
      this.setBackground(UIManager.getColor("List.textBackground"));
      this.setText(className);
      this.setSelected(enabledAspects.contains(className));
      return this;
    }
  }
}
