/**
 * [BSD Open Source License]
 *
 * Copyright (c) 2006, Mattias Jiderhamn, Expert Systems
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or othermaterials provided with the distribution.
 *
 * Neither the name of Expert Systems nor the names of its contributors may
 * be usedto endorse or promote products derived from this software without
 * specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package se.expertsystem.intellij.aop.aspectj;

import com.intellij.openapi.diagnostic.Logger;
import com.intellij.openapi.options.Configurable;
import com.intellij.openapi.project.Project;
import com.intellij.openapi.ui.Messages;
import com.intellij.openapi.util.InvalidDataException;
import com.intellij.openapi.util.WriteExternalException;
import com.intellij.openapi.util.JDOMExternalizer;
import org.jdom.Element;
import se.expertsystem.intellij.PostProcessProjectComponent;

import javax.swing.*;
import java.util.*;


/**
 * Project component for AspectJ plugin
 * @author Mattias Jiderhamn, mattias.jiderhamn@expertsystems.se
 */
public class AspectJProjectComponent extends PostProcessProjectComponent implements Configurable {

  private static final Logger LOG = Logger.getInstance(AspectJProjectComponent.class.getName());

  /** Handle to configuration dialog, while opened */
  private AspectJConfiguration configuration;

  /** Map with the location of the class files of all known aspects. Class name -> class file URL */
  private SortedMap/*<String,String>*/ aspectLocations = new TreeMap();

  /** Set with class names of all enabled aspects */
  private Set/*<String>*/ enabledAspects = new HashSet();
  
  private boolean noInline;

  /**
   * IoC constructor
   * @param project
   */
  public AspectJProjectComponent(Project project) {
    this.project = project;
    this.postCompiler = new AspectJWeaver(this);
  }

  public String getComponentName() {
    return AspectJProjectComponent.class.getName();
  }

  public void projectOpened() {
  }

  public void projectClosed() {
    // When project closes, disable to unwrap compiler
    changeTo(false);
  }

  public void initComponent() {
  }

  public void disposeComponent() {
  }

  /**
   * Called when component is enabled from the Build menu
   */
  public void changeTo(boolean activate) {
    super.changeTo(activate);
    // When enabled for the first time, give a little tip
    if(isActivated && aspectLocations.isEmpty()) {
      Messages.showInfoMessage("If aspects are not compiled by this project,\n" +
          "you must search for them in the project settings.",
          "Search for aspects");
    }
  }

  protected void configLoaded(boolean activated) {
    super.changeTo(activated);
    // Do NOT give tip about the lack of aspects during load, since IntelliJ 7 hangs!
  }

  /**
   * Read on/off state from IWS file
   */
  public void readExternal(Element element) throws InvalidDataException {
    // Read aspects first, since they are needed if component is activated
    Map knownAspects = getAspectLocations();
    List aspectElems = element.getChildren("aspect");
    for(Iterator iter = aspectElems.iterator(); iter.hasNext(); ) {
      Element aspectElem = (Element) iter.next();
      final String className = aspectElem.getAttributeValue("className");
      knownAspects.put(className, aspectElem.getText());
      if("true".equals(aspectElem.getAttributeValue("enabled")))
        enabledAspects.add(className);
    }
    LOG.info("Aspects found in project workspace: " + knownAspects);
    LOG.info("Aspects enabled in project workspace: " + enabledAspects);
    super.readExternal(element); // Read on/off state
    noInline = JDOMExternalizer.readBoolean(element, "noInline");
  }

  /**
   * Persists state in IWS file
   */
  public void writeExternal(Element element) throws WriteExternalException {
    super.writeExternal(element); // Write on/off state
    if(noInline)
      JDOMExternalizer.write(element, "noInline", noInline);
    // Write all known aspects to project file
    Map knownAspects = getAspectLocations();
    LOG.info("Writing known aspects to project workspace: " + knownAspects);
    // LOG.info("Writing enabled aspects to project workspace: " + enabledAspects);
    for(Iterator iter = knownAspects.entrySet().iterator(); iter.hasNext(); ) {
      Map.Entry entry = (Map.Entry) iter.next();
      Element aspectElem = new Element("aspect");
      final String className = (String) entry.getKey();
      aspectElem.setAttribute("className", className);
      aspectElem.setText((String) entry.getValue());
      if(enabledAspects.contains(className))
        aspectElem.setAttribute("enabled", "true");
      element.addContent(aspectElem);
    }
  }

  /** Get the locations of all known aspects */
  public SortedMap/*<String,String>*/ getAspectLocations() {
    return aspectLocations;
  }

  /** Get the set of enabled aspects */
  public Set/*<String>*/ getEnabledAspects() {
    return enabledAspects;
  }

  public boolean isNoInline() {
    return noInline;
  }

  /** Component display name*/
  public String getDisplayName() {
    return "AspectJ weaver";
  }

  /** Setting menu icon */
  public Icon getIcon() {
    return new ImageIcon(this.getClass().getResource("logo.gif"));
  }

  public String getHelpTopic() {
    return null;
  }

  /** Open the setting dialog */
  public JComponent createComponent() {
    configuration = new AspectJConfiguration(project, isActivated, getAspectLocations(), enabledAspects);
    return configuration;
  }

  /** Check if settings has been modified in dialog */
  public boolean isModified() {
    return ! (isActivated == configuration.isEnabled() &&
              aspectLocations.equals(configuration.getAspectLocations()) &&
              enabledAspects.equals(configuration.getEnabledAspects()));
  }

  /** Apply new configuration from dialog */
  public void apply() {
    LOG.info("New configuration applied");
    super.changeTo(configuration.isEnabled());
    this.aspectLocations.clear();
    this.aspectLocations.putAll(configuration.getAspectLocations());
    this.enabledAspects.clear();
    this.enabledAspects.addAll(configuration.getEnabledAspects());
    LOG.debug("  aspectLocations: " + aspectLocations);
    LOG.debug("  enabledAspects: " + enabledAspects);
  }

  public void reset() {
    LOG.debug("Config reset");
  }

  /** Clean up after configuration dialog close */
  public void disposeUIResources() {
    configuration = null;
  }
}