/**
 * [BSD Open Source License]
 *
 * Copyright (c) 2006, Mattias Jiderhamn, Expert Systems
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or othermaterials provided with the distribution.
 *
 * Neither the name of Expert Systems nor the names of its contributors may
 * be usedto endorse or promote products derived from this software without
 * specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package se.expertsystem.intellij.aop.aspectj;

import com.intellij.openapi.compiler.CompileContext;
import com.intellij.openapi.compiler.CompilerMessageCategory;
import com.intellij.openapi.diagnostic.Logger;
import com.intellij.openapi.progress.ProgressIndicator;
import com.intellij.openapi.vfs.VfsUtil;
import com.intellij.openapi.vfs.VirtualFile;
import com.intellij.openapi.vfs.LocalFileSystem;
import com.intellij.openapi.module.Module;
import com.intellij.openapi.roots.ModuleRootManager;
import com.intellij.util.ActionRunner;
import org.aspectj.apache.bcel.classfile.JavaClass;
import org.aspectj.bridge.AbortException;
import org.aspectj.bridge.IMessage;
import org.aspectj.bridge.IMessageHandler;
import org.aspectj.bridge.ISourceLocation;
import org.aspectj.util.FileUtil;
import org.aspectj.weaver.IClassFileProvider;
import org.aspectj.weaver.IUnwovenClassFile;
import org.aspectj.weaver.IWeaveRequestor;
import org.aspectj.weaver.ReferenceType;
import org.aspectj.weaver.bcel.BcelWeaver;
import org.aspectj.weaver.bcel.BcelWorld;
import org.aspectj.weaver.bcel.UnwovenClassFile;
import org.aspectj.weaver.bcel.Utility;
import se.expertsystem.intellij.PluginUtils;
import se.expertsystem.intellij.PostProcessCompiler;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.*;

/**
 * Instrumenting compiler that weaves aspects into class files using AspecJ.
 * @author Mattias Jiderhamn, mattias.jiderhamn@expertsystems.se
 */
public class AspectJWeaver extends PostProcessCompiler {

  private static final Logger LOG = Logger.getInstance(AspectJWeaver.class.getName());

  private final AspectJProjectComponent projectComponent;

  public AspectJWeaver(AspectJProjectComponent projectComponent) {
    this.projectComponent = projectComponent;
  }

  /** Get the processing items this compiler should process. Will return all the files
   * recompiled by the (wrapped) Java compiler.*/
  public final ProcessingItem[] getProcessingItems(final CompileContext compileContext) {
    final List/*<ModuleTaggedProcessingItem>*/ items = getModuleTaggedProcessingItems(compileContext);

    LOG.debug("AspectJWeaver.getProcessingItems found " + items.size() + " items");
    return (ProcessingItem[]) items.toArray(new ProcessingItem[]{});
  }

  public ProcessingItem[] process(final CompileContext compileContext, final ProcessingItem[] processingItems) {

    // Correctly weaved
    // final List processedItems = new ArrayList();

    compileContext.addMessage(CompilerMessageCategory.INFORMATION,
        "AspectJ: Weaving " + processingItems.length + " items(s)", null, -1, -1);
    LOG.info("AspectJ: Weaving " + processingItems.length + " item(s)...");

    // Sort files per module
    final Map filesPerModule = sortItemsPerModule(processingItems);

    // Loop modules and weave
    for(Iterator iter = filesPerModule.entrySet().iterator(); iter.hasNext(); ) {
      Map.Entry entry = (Map.Entry) iter.next();
      Module module = (Module) entry.getKey();
      List files = (List) entry.getValue();
      weave(compileContext, module, files);
    }

    return processingItems;
    // Return only correctly processed items
    // return (ProcessingItem[]) processedItems.toArray(new ProcessingItem[]{});
  }

  /** Weave re-compiled files in single module */
  private void weave(final CompileContext compileContext, final Module module,
                     final List/*<ProcessingItem>*/ processingItems /*, final List/*<ProcessingItem> processedItems*/) {
    final ProgressIndicator progressIndicator = compileContext.getProgressIndicator();

    try {
      // Run in an IDEA read action since we access VirtualFiles..
      ActionRunner.runInsideReadAction(
          new ActionRunner.InterruptibleRunnable() {
            public void run() throws Exception {
              LOG.debug("Searching for aspects in project");


              BcelWorld world = createWorld(module, compileContext);

              world.setXnoInline(projectComponent.isNoInline());
              final Map aspectLocations = projectComponent.getAspectLocations();
              Set enabledAspects = projectComponent.getEnabledAspects();
              /*
              for(Iterator iter = enabledAspects.iterator(); iter.hasNext(); ) {
                String aspectClassName = (String) iter.next();
                String url = (String) aspectLocations.get(aspectClassName);
                if(url != null) {
                  ReferenceType aspect = AspectFinder.readAspect(url, world);
                  if(aspect.isAspect())
                    world.getCrosscuttingMembersSet().addOrReplaceAspect(aspect);
                }
              }
              */

              Set disabledAspectLocations = new HashSet();
              for(Iterator iter = aspectLocations.entrySet().iterator(); iter.hasNext(); ) {
                Map.Entry entry = (Map.Entry) iter.next();
                String className = (String) entry.getKey();
                String url = (String) entry.getValue();
                if(enabledAspects.contains(className)) {
                  ReferenceType aspect = AspectFinder.readAspect(url, world);
                  if(aspect != null && aspect.isAspect())
                    world.getCrosscuttingMembersSet().addOrReplaceAspect(aspect);
                  else // Not found or no longer aspect (may be readded after weaving below)
                    aspectLocations.remove(className);
                }
                else
                  disabledAspectLocations.add(url);
              }

              // Sort possibly adviced classes/active aspects and disabled aspects
              final List/*<UnwovenClassFile>*/ filesToWeave = new ArrayList();
              final List/*<UnwovenClassFile>*/ disabledAspectsToWeave = new ArrayList();
              for(Iterator/*<ProcessingItem>*/ iter = processingItems.iterator(); iter.hasNext(); ) {
                File fileToWeave = VfsUtil.virtualToIoFile(((ProcessingItem)iter.next()).getFile());
                byte[] bytes = FileUtil.readAsByteArray(fileToWeave);
                UnwovenClassFile unwovenClassFile = new UnwovenClassFile(fileToWeave.getAbsolutePath(),bytes);
                if(! disabledAspectLocations.contains(PluginUtils.toUrlString(fileToWeave))) {
                  filesToWeave.add(unwovenClassFile);
                  LOG.debug("Queue for weave: " + fileToWeave);
                }
                else {
                  disabledAspectsToWeave.add(unwovenClassFile);
                  LOG.info("Aspect '" + fileToWeave + "' disabled, queued for separate weave");
                }
                // processedItems.add(processingItem); // Remember as processed
              }

              BcelWeaver weaver = new BcelWeaver(world);
              Map unwovenFiles = addForWeaving(weaver, filesToWeave, compileContext);

              int totalFiles = unwovenFiles.size();
              LOG.debug("Weaving");
              progressIndicator.setText2("AspectJ: Weaving module '" + module.getName() + "'");
              progressIndicator.setIndeterminate(false);

              WeaveRequestor weaveRequestor = new WeaveRequestor(unwovenFiles, aspectLocations, enabledAspects,
                  compileContext, world, progressIndicator, totalFiles);
              weaver.prepareForWeave();
              weaver.weave(new ClassFileProvider(weaveRequestor, filesToWeave));

              if(! disabledAspectsToWeave.isEmpty()) {
                world = createWorld(module, compileContext);
                weaver = new BcelWeaver(world);
                totalFiles = disabledAspectsToWeave.size();
                LOG.debug("Weaving disabled aspects: " + disabledAspectsToWeave);
                progressIndicator.setText2("AspectJ: Weaving disabled aspects in module '" + module.getName() + "'");
                progressIndicator.setIndeterminate(false);

                unwovenFiles = addForWeaving(weaver, disabledAspectsToWeave, compileContext);
                weaveRequestor = new WeaveRequestor(unwovenFiles, aspectLocations, enabledAspects,
                    compileContext, world, progressIndicator, totalFiles);
                weaver.prepareForWeave();
                weaver.weave(new ClassFileProvider(weaveRequestor, disabledAspectsToWeave));
              }

              progressIndicator.setIndeterminate(true);
            }

          });
    } catch (Exception e) {
      LOG.error(e);
      compileContext.addMessage(CompilerMessageCategory.ERROR, e.getMessage(), null, -1, -1);
    }
  }

  private static BcelWorld createWorld(final Module module, final CompileContext compileContext) {
    final BcelWorld world = new BcelWorld();
    world.setBehaveInJava5Way(PluginUtils.isJava5(module));
    addClasspath(world, module); // Add module classpath to world
    // Add outputs of all modules, since they may contain declared parents
    Module[] modules = PluginUtils.getProjectModules(module.getProject());
    for(int i = 0; i < modules.length; i++) {
      List/*<File>*/ outputPaths = PluginUtils.getModuleOutputs(modules[i]);
      for(Iterator/*<File>*/ iter = outputPaths.iterator(); iter.hasNext(); )
        world.addPath(((File)iter.next()).getAbsolutePath());
    }
    world.setMessageHandler(new IMessageHandler() {

      private Set/*<IMessage.Kind>*/ ignored = new HashSet/*<IMessage.Kind>*/();

      public boolean handleMessage(IMessage iMessage) throws AbortException {
        if(isIgnoring(iMessage.getKind())) {
          ; // Don't do anything
        }
        else if(iMessage.isError() || iMessage.isWarning()) {
          CompilerMessageCategory category = iMessage.isError() ?
              CompilerMessageCategory.ERROR :
              CompilerMessageCategory.WARNING;
          LOG.debug("iMessage.getMessage(): " + iMessage.getMessage());
          LOG.debug("iMessage.getDetails(): " + iMessage.getDetails());
          if(iMessage.isError())
            LOG.debug(new Throwable("Stacktrace"));
          final ISourceLocation sourceLocation = iMessage.getSourceLocation();
          if(sourceLocation != null) {
            VirtualFile sourceFile = null;
            // File messageSourceFile = sourceLocation.getSourceFile();
            final ModuleRootManager rootManager = ModuleRootManager.getInstance(module);
            final VirtualFile[] sourceRoots = rootManager.getSourceRoots();
            for(int i = 0; i < sourceRoots.length; i++) {
              final VirtualFile fileInSource = VfsUtil.findRelativeFile(sourceLocation.getSourceFile().getPath(), sourceRoots[i]);
              // LOG.debug("VfsUtil.findRelativeFile: " + fileInSource);
              if(fileInSource != null && fileInSource.exists()) {
                sourceFile = fileInSource;
                break;
              }

              // File srcFile = new File(VfsUtil.virtualToIoFile(sourceRoots[i]), messageSourceFile.getPath());
            }

            LOG.debug("sourceFile: " + sourceFile);
            // PluginUtils.toUrlString(messageSourceFile)) gives "file:/" while "file://" is required
            // VfsUtil.findFileByURL(PluginUtils.toURL(messageSourceFile)).getUrl(), // OK
            compileContext.addMessage(category, iMessage.getMessage(),
                (sourceFile != null) ? sourceFile.getUrl() : null,
                sourceLocation.getLine(), sourceLocation.getColumn());
          }
          else
            compileContext.addMessage(category, iMessage.getMessage(), null, -1, -1);

          // User may have used another JAR than aspectjrt.jar - we could warn them by using this check
          // if("can't find critical required type org.aspectj.lang.JoinPoint$StaticPart".equals(iMessage.getMessage()))
          //  ...;
        }
        else {
          // compileContext.addMessage(CompilerMessageCategory.INFORMATION, iMessage.getMessage(), null, -1, -1);
        }
        return true;
      }

      public boolean isIgnoring(IMessage.Kind kind) {
        return ignored.contains(kind);
      }

      public void dontIgnore(IMessage.Kind kind) {
        ignored.remove(kind);
      }

      public void ignore(IMessage.Kind kind) {
        ignored.add(kind);
      }
    });
    return world;
  }

  /**
   * Utility method to add the module classpath to the BcelWorld.
   */
  private static void addClasspath(final BcelWorld world, final Module module) {
    try {
      List classpath = PluginUtils.getModuleClasspath(module, true);

      for(Iterator iter = classpath.iterator(); iter.hasNext(); ) {
        File cpEntry = (File) iter.next();
        world.addPath(cpEntry.getAbsolutePath());
        LOG.debug(cpEntry + " added to weaver");
      }
    }
    catch(IOException ioex) {
      LOG.error(ioex);
    }
  }

  /**
   * Add list of files as classes to weave by weaver. Will return a Map with class name as key and UnwovenClassFile as value.
   * This map will later be used to identify where the file should be written.
   * @param weaver
   * @param filesToWeave
   * @param compileContext
   * @return
   */
  private static Map addForWeaving(final BcelWeaver weaver, final List filesToWeave, final CompileContext compileContext) {
    final Map/*<String,UnwovenClassFile>*/ unwovenFiles = new HashMap();
    for(Iterator/*<UnwovenClassFile>*/ iter = filesToWeave.iterator(); iter.hasNext(); ) {
      UnwovenClassFile unwovenClassFile = (UnwovenClassFile) iter.next();
      weaver.addClassFile(unwovenClassFile, false);
      if(unwovenFiles.put(unwovenClassFile.getClassName(), unwovenClassFile) != null) {
        LOG.error("Class already inserted " + unwovenClassFile.getClassName());
        compileContext.addMessage(CompilerMessageCategory.WARNING, "Cannot weave multiple classes with same name: " + unwovenClassFile.getClassName(), null, -1, -1);
      }
    }
    return unwovenFiles;
  }

  public String getDescription() {
    return "AspectJ weaver";
  }

  private static class ClassFileProvider implements IClassFileProvider {
    private final WeaveRequestor weaveRequestor;
    private final List filesToWeave;

    public ClassFileProvider(WeaveRequestor weaveRequestor, List filesToWeave) {
      this.weaveRequestor = weaveRequestor;
      this.filesToWeave = filesToWeave;
    }

    public boolean isApplyAtAspectJMungersOnly() {
      return false;
    }

    public Iterator getClassFileIterator() {
      return filesToWeave.iterator();
    }

    public IWeaveRequestor getRequestor() {
      return weaveRequestor;
    }
  }

  /**
   * Inner class to handle writing of weaved classes and progress indication.
   */
  private static class WeaveRequestor implements IWeaveRequestor {
    /** Map of unwoven files, so that we can see what files have changed, where to write them */
    private final Map unwovenFiles;

    /** Map of aspect locations, so we can remember classes that turns out to be aspects */
    private final Map/*<String,String>*/ aspectLocations;

    /** Set of enabled aspects, so newly found aspects can be enabled by default */
    private final Set/*<String>*/ enabledAspects;

    /** Compile context for error messages */
    private final CompileContext compileContext;

    /** Weaving world */
    private final BcelWorld world;

    /** Progress indicator */
    private final ProgressIndicator progressIndicator;

    /** No of total files, for progress indication */
    private final int totalFiles;

    /** Current file, for progress indication */
    private int count = 0;

    public WeaveRequestor(Map unwovenFiles, Map aspectLocations, Set enabledAspects,
                          CompileContext compileContext, BcelWorld world, ProgressIndicator progressIndicator, int totalFiles) {
      this.progressIndicator = progressIndicator;
      this.totalFiles = totalFiles;
      this.unwovenFiles = unwovenFiles;
      this.compileContext = compileContext;
      this.aspectLocations = aspectLocations;
      this.enabledAspects = enabledAspects;
      this.world = world;
    }

    public void acceptResult(IUnwovenClassFile result) { // Woven class
      PluginUtils.setProgress(progressIndicator, totalFiles, ++count);
      final String className = result.getClassName();
      LOG.debug("Done weaving " + className + "; writing to disk.");
      progressIndicator.setText(className);

      // NOTE!!! result.getFilename() does not contain path info!
      // Therefore, we must find path of unwoven file

      UnwovenClassFile unwovenFile = (UnwovenClassFile) unwovenFiles.get(className);
      if(unwovenFile == null) { // Unknown class
        final int dollarIndex = className.indexOf('$'); // Inner class created by AspectJ
        if(dollarIndex != -1) {
          // Construct filename from outer class filename
          final String parentClassName = className.substring(0, dollarIndex);
          final String parentFileName = ((UnwovenClassFile) unwovenFiles.get(parentClassName)).getFilename();
          final String fileName = parentFileName.substring(0, parentFileName.indexOf(".class")) +
              '$' + className.substring(dollarIndex + 1)+ ".class";
          writeResult(fileName, result); // Write weaved bytes to disk
        }
        else
          throw new RuntimeException("Unwoven file not found for class " + className);
      }
      else if(result != unwovenFile) { // File was actually weaved
        writeResult(unwovenFile.getFilename(), result); // Write weaved bytes to disk

        // If new aspect, add to set of known aspects
        if(! aspectLocations.containsKey(className)) {
          JavaClass jc = Utility.makeJavaClass(result.getFilename(), result.getBytes());
          ReferenceType type = AspectFinder.getResolvedType(jc, world);
          if(type.isAspect()) {
            LOG.debug("Found new aspect: " + type.getName());
            aspectLocations.put(type.getName(),
                PluginUtils.toUrlString(new File(unwovenFile.getFilename()))); // Remember as aspect
            enabledAspects.add(type.getName()); // Make new aspect enabled by default
          }
        }
      }
    }

    /** Write woven class to file */
    private void writeResult(String fileName, IUnwovenClassFile result) {
      LOG.debug(result.getClassName() + " was woven, saving to " + fileName);
      try {
        FileOutputStream fos = new FileOutputStream(fileName);
        fos.write(result.getBytes());
        fos.flush();
        fos.close();
      }
      catch(IOException ioex) {
        LOG.error(ioex);
        compileContext.addMessage(CompilerMessageCategory.ERROR, ioex.getMessage(), fileName, -1, -1);
      }
    }

    public void processingReweavableState() {
      progressIndicator.setText2("AspectJ: Processing reweavable state");
    }
    public void addingTypeMungers() {
      progressIndicator.setText2("AspectJ: Adding type munger");
    }
    public void weavingAspects() {
      progressIndicator.setText2("AspectJ: Weaving aspects");
    }
    public void weavingClasses() {
      progressIndicator.setText2("AspectJ: Weaving classes");
    }
    public void weaveCompleted() {
      progressIndicator.setText2("AspectJ: Done");
    }
  }


}