/**
 * [BSD Open Source License]
 *
 * Copyright (c) 2006, Mattias Jiderhamn, Expert Systems
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or othermaterials provided with the distribution.
 *
 * Neither the name of Expert Systems nor the names of its contributors may
 * be usedto endorse or promote products derived from this software without
 * specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package se.expertsystem.intellij;

import com.intellij.openapi.application.ApplicationInfo;
import com.intellij.openapi.application.ApplicationManager;
import com.intellij.openapi.diagnostic.Logger;
import com.intellij.openapi.module.Module;
import com.intellij.openapi.module.ModuleManager;
import com.intellij.openapi.progress.ProgressIndicator;
import com.intellij.openapi.progress.ProgressManager;
import com.intellij.openapi.project.Project;
import com.intellij.openapi.roots.ModuleRootManager;
import com.intellij.openapi.roots.OrderEnumerator;
import com.intellij.openapi.util.io.FileUtil;
import com.intellij.openapi.vfs.VfsUtil;
import com.intellij.openapi.vfs.VirtualFile;
import com.intellij.util.ActionRunner;

import java.io.File;
import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.net.MalformedURLException;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URL;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * Class containing useful utility methods for IntelliJ IDEA plugins.
 * @author Mattias Jiderhamn
 */
public abstract class PluginUtils {
  
  private static final Logger LOG = Logger.getInstance(PluginUtils.class.getName());

  private static final Class[] NO_PARAMETERS = new Class[] {};
  
  private static final Pattern JDK_VERSION_PATTERN = Pattern.compile("(java|openjdk)\\s+version\\s+\"(\\d+)\\.(\\d+)");  

  /**
   * Find classpath for module. Returns list of <code>File</code>s.
   * @param project Project to get classpath for.
   * @param includeJdk If false, the JDK JARs will be excluded from the output.
   * @return
   */
  public static List/*<File>*/ getProjectClasspath(final Project project, boolean includeJdk) throws IOException {
    List/*<List>*/ output = new ArrayList(); 
    Module[] modules = getProjectModules(project);
    for(int i = 0; i < modules.length; i++) {
      output.addAll(getModuleClasspath(modules[i], includeJdk));
    }
    return output;
  }

  /**
   * Find classpath for module. Returns list of <code>File</code>s.
   * @param module Module to get classpath for.
   * @param includeJdk If false, the JDK JARs will be excluded from the output.
   * @return
   */
  public static List/*<File>*/ getModuleClasspath(final Module module, final boolean includeJdk) throws IOException {
    String jdkPath = getJdkHomePath(module);
    final File jdkRtPath = new File(jdkPath);
    final List classpath = new ArrayList();
    for(VirtualFile virtualFile : OrderEnumerator.orderEntries(module).recursively().exportedOnly().getPathsList().getVirtualFiles()) {
      File f = VfsUtil.virtualToIoFile(virtualFile);
      if(f.exists()) {
        if(includeJdk || ! FileUtil.isAncestor(jdkRtPath, f, true)) {
          classpath.add(f);
        }
      }
    }
                    
    // LOG.debug("Classpath for module '" + module.getName() + "': " + classpath);
    return classpath;
  }

  /**
   * Find output path for module. Returns list of <code>File</code>s.
   * @param module Module to get output path for.
   */
  public static List/*<File>*/ getModuleOutputs(final Module module) {
    List classpath = new ArrayList();

    VirtualFile compilerOutputPath = getCompilerOutputPath(module);
    if(compilerOutputPath != null && compilerOutputPath.exists()) {
      File outputPath = VfsUtil.virtualToIoFile(compilerOutputPath);
      if(outputPath.exists())
        classpath.add(outputPath);
    }
    VirtualFile compilerOutputPathForTests = getCompilerOutputPathForTests(module);
    if(compilerOutputPathForTests != null && compilerOutputPathForTests.exists()) {
      File testOutputPath = VfsUtil.virtualToIoFile(compilerOutputPathForTests);
      if(testOutputPath.exists())
        classpath.add(testOutputPath);
    }
                    
    // LOG.debug("Output path for module '" + module.getName() + "': " + classpath);
    return classpath;
  }
  
  public static VirtualFile getCompilerOutputPath(final Module module) {
    return getModulePath("getCompilerOutputPath", module);
  }

  public static VirtualFile getCompilerOutputPathForTests(final Module module) {
    return getModulePath("getCompilerOutputPathForTests", module);
  }

  private static VirtualFile getModulePath(String methodName, final Module module) {
    Class clazz = ModuleRootManager.class; // Assue IDEA < 7 or earlier 
    try {
      clazz = Class.forName("com.intellij.openapi.roots.CompilerModuleExtension"); // IDEA v 8
    }
    catch(ClassNotFoundException cnfex) {
      // Do nothing; IDEA < 7 or earlier
    }

    try {
      final Method instanceMethod = clazz.getDeclaredMethod("getInstance", new Class[]{Module.class});
      Object instance = instanceMethod.invoke(null, new Object[] {module});
      return (VirtualFile) clazz.getMethod(methodName, NO_PARAMETERS).invoke(instance, null);
    }
    catch (NoSuchMethodException e) {
      throw new RuntimeException(e);
    }
    catch (IllegalAccessException e) {
      throw new RuntimeException(e);
    }
    catch (InvocationTargetException e) {
      throw new RuntimeException(e);
    }
  }

  /**
   * Find classpath for this plugin. Will return list of <code>File</code>s.
   */
  public static List/*<File>*/ getPluginClasspath() {
    /*
      Will be loaded by a plugin classloader that has an ArrayList
      of URLs. (For IDEA 4.5.4 this is in com.intellij.ide.plugins.a.a.c()
    */

    ClassLoader cl = PluginUtils.class.getClassLoader();
    // LOG.debug("Plugin classloader is: " + cl.getClass());
    Method[] methods = cl.getClass().getMethods();
    for(int i = 0; i < methods.length; i++) {
      Method method = methods[i];
      if(List.class.isAssignableFrom(method.getReturnType())) {
        try {
          List classpath = new ArrayList();
          List urls = (List) method.invoke(cl, (Object[])null);
          for(Iterator iter = urls.iterator(); iter.hasNext(); ) {
            URL url = (URL) iter.next();
            classpath.add(toFile(url));
          }
          return classpath;
        }
        catch(Exception ex) {
          LOG.error(ex);
        }
      }
    }
    return null;
  }

  /** Convert URL string to <code>File</code> */
  public static File toFile(String url) throws MalformedURLException {
    return toFile(new URL(url));
  }

  /** Convert <code>URL</code> to <code>File</code> */
  public static File toFile(URL url) {
    // if("file".equals(url.getProtocol()))
    //   return new File(url.getFile());
    
    try {
      // String decoded = URLDecoder.decode(url.toString().substring(6));
      String s = url.toString().replaceAll(" ", "%20");
      if(s.startsWith("file:") && ! s.startsWith("file:/")) // Fix Windows path name
        s = "file:/" + s.substring("file:".length());
      return new File(new URI (s));
    }
    catch(URISyntaxException usex) { // Should not happen
      LOG.error(usex);
      throw new IllegalArgumentException("Illegal URL: " + url);
    }
  }

  /** Convert <code>File</code> to URL string */
  public static String toUrlString(File file) {
    return toURL(file).toString();
  }

  /** Convert <code>File</code> to <code>URL</code> */
  public static URL toURL(File file) {
    try {
      return file.toURI().toURL();
    }
    catch(MalformedURLException muex) {
      LOG.error(muex);
      return null;
    }
  }

  /** Given a content path and a (source or class) file path, resolve the package name of the file */
  public static String resolvePackageName(String rootPath, String classPath) {
    String relativePath = classPath.substring(rootPath.length() + 1);
    int lastSlash = relativePath.lastIndexOf('/');
    return (lastSlash != -1) ? relativePath.substring(0, lastSlash).replace('/', '.') : null;
  }

  public static String getJdkHomePath(Module module) {
    return getModuleRootString(module, "getHomePath");
  }

  public static String getJdkVersionString(Module module) {
    return getModuleRootString(module, "getVersionString");
  }

  private static String getModuleRootString(Module module, String methodName) {
    final ModuleRootManager moduleRootManager = ModuleRootManager.getInstance(module);
    try {
      Object jdk;
      try {
        jdk = ModuleRootManager.class.getMethod("getJdk", NO_PARAMETERS).invoke(moduleRootManager, null); // IDEA v 7 and earlier
      }
      catch (NoSuchMethodException e) {
        jdk = ModuleRootManager.class.getMethod("getSdk", NO_PARAMETERS).invoke(moduleRootManager, null); // IDEA v 8
      }
      if(jdk == null)
        throw new RuntimeException("Module " + module.getName() + " has no JDK!");
      return (String) jdk.getClass().getMethod(methodName, NO_PARAMETERS).invoke(jdk, null);
    }
    catch (IllegalAccessException e) {
      throw new RuntimeException(e);
    }
    catch (InvocationTargetException e) {
      throw new RuntimeException(e);
    }
    catch (NoSuchMethodException e) {
      throw new RuntimeException(e);
    }
  }

  /**
   * Check if project uses Java 5
   * @return true if any module uses a Java 1.5 JDK, otherwise false
   */
  public static boolean isJava5(Project project) {
    Module[] modules = getProjectModules(project);
    for(int i = 0; i < modules.length; i++) {
      if(isJava5(modules[i]))
        return true;
    }
    return false;
  }

  /**
   * Check if module uses Java 5 or later. Tested with 1.6 RC (true).
   * Thanks to Eugene N Dzhurinsky for OpenJDK support. 
   * @return true if module uses a Java 1.5 or later JDK, otherwise false.
   */
  public static boolean isJava5(Module module) {
    final String versionString = getJdkVersionString(module);  // Example: java version "1.5.0_07"
    LOG.debug("Detected JDK: " + versionString);
    final Matcher versionMatcher = JDK_VERSION_PATTERN.matcher(versionString);
    if(versionMatcher.find() && 
       "1".equals(versionMatcher.group(2)) && versionMatcher.group(3).charAt(0) < '5') { // JDK 1.x < 1.5
      LOG.debug("Pre 1.5 JDK: major " + versionMatcher.group(2)+ " , minor " + versionMatcher.group(3));
      return false;
    }
    else
      return true; // Assume JDK 1.5+
  }

  /////////////////////////////////////////////////////////////////////////////
  // Methods for progress indication
  /////////////////////////////////////////////////////////////////////////////

  /**
   * Update fraction of progress indication by performing fraction calculation
   * @param total Total number of items to process
   * @param current Current item processed.
   */
  public static void setProgress(ProgressIndicator progressIndicator, int total, int current) {
    progressIndicator.setFraction(1D / ((double)total/(double)current));
  }

  /**
   * (Helper method to handle change in Open API. Will use reflection to execute Runnable
   * with progress bar.)
   */
  public static void runProcessWithProgressSynchronously(Runnable runnable, String title,
                                                         boolean cancelable, Project project) {
    String currentBuildNo = ApplicationInfo.getInstance().getBuildNumber();
    if("5131".compareTo(currentBuildNo) <= 0) { // IntelliJ 6 (6.0 = 5581)
      runProcessWithProgressSynchronously(ProgressManager.getInstance(),
          runnable, title, cancelable, project);
    }
    else { // Older IntelliJ
      runProcessWithProgressSynchronously(ApplicationManager.getApplication(),
          runnable, title, cancelable, project);
    }
  }

  private static void runProcessWithProgressSynchronously(Object o, Runnable runnable, String title,
                                                         boolean cancelable, Project project) {
    try {
      Method method = o.getClass().getMethod("runProcessWithProgressSynchronously",
          new Class[] {Runnable.class, String.class, boolean.class, Project.class});
      method.invoke(o, new Object[] {runnable, title, cancelable ? Boolean.TRUE : Boolean.FALSE, project});
    }
    catch(Exception ex) {
      LOG.error(ex);
    }
  }
  
  /////////////////////////////////////////////////////////////////////////////
  // Methods and classes for finding modules
  /////////////////////////////////////////////////////////////////////////////

  public static Module[] getProjectModules(final Project project) {
    final ProjectModuleFinder projectModuleFinder = new ProjectModuleFinder(project);
    try {
      ActionRunner.runInsideReadAction(projectModuleFinder);
    } catch (Exception e) {
      LOG.error(e);
    }
    
    return projectModuleFinder.getModules();
  }

  /** Runnable used inside read action to find the modules of a project */
  private static class ProjectModuleFinder implements ActionRunner.InterruptibleRunnable {
    /** The project to look for modules in */
    private final Project project;

    /** The modules found*/
    private Module[] modules;
    
    
    public ProjectModuleFinder(Project project) {
      this.project = project;
    }

    /** The method that actually finds the modules */
    public void run() {
      modules = ModuleManager.getInstance(project).getModules();
    }

    /** Get the found moules */
    public Module[] getModules() {
      return modules;
    }
  }
}