/**
 * [BSD Open Source License]
 *
 * Copyright (c) 2006, Mattias Jiderhamn, Expert Systems
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or othermaterials provided with the distribution.
 *
 * Neither the name of Expert Systems nor the names of its contributors may
 * be usedto endorse or promote products derived from this software without
 * specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package se.expertsystem.intellij;

import com.intellij.openapi.compiler.*;
import com.intellij.openapi.module.Module;
import com.intellij.openapi.vfs.VirtualFile;
import com.intellij.openapi.vfs.VfsUtil;
import com.intellij.openapi.diagnostic.Logger;
import com.intellij.openapi.util.Key;
import com.intellij.util.ActionRunner;

import java.util.*;
import java.io.IOException;
import java.io.DataInputStream;
import java.io.DataInput;

/**
 * Base class for AOP weaving compiler. Will process all files compiled by the standard Java compiler.
 * There files are kept in a list updated by the WrappingCompiler.
 * @author Mattias Jiderhamn
 */
public abstract class PostProcessCompiler implements ClassInstrumentingCompiler {

  private static Logger LOG = Logger.getInstance(PostProcessCompiler.class.getName());

  /** Key used for recompiled classes in CompileContext  */
  public static final Key<List<RecentlyCompiledSink.CompiledItem>> COMPILED_CLASSES = new CrossClassLoaderKey<List<RecentlyCompiledSink.CompiledItem>>("Compiled classes queued for post-process");
  
  
  public static List<RecentlyCompiledSink.CompiledItem> getRecentlyCompiled(final CompileContext compileContext) {
    List<RecentlyCompiledSink.CompiledItem> list = compileContext.getUserData(COMPILED_CLASSES);
    if(list == null) {
      list = new ArrayList<RecentlyCompiledSink.CompiledItem>();
      compileContext.putUserData(COMPILED_CLASSES, list);
  	}
    return list;
  }

  protected List/*<ProcessingItem>*/ getQueuedProcessingItems(final CompileContext compileContext) {
    final List items = new ArrayList(); // List of items to return
    try {
      ActionRunner.runInsideReadAction(new ActionRunner.InterruptibleRunnable() {
        public void run() {
          for(Iterator<RecentlyCompiledSink.CompiledItem> iter = getRecentlyCompiled(compileContext).iterator(); iter.hasNext(); ) {
            RecentlyCompiledSink.CompiledItem outputItem = iter.next();
            String packageName = PluginUtils.resolvePackageName(outputItem.getOutputRoot(), outputItem.getOutputPath());
            // (Requires read action)
            VirtualFile vf = VfsUtil.findFileByURL(
                VfsUtil.convertToURL(VfsUtil.pathToUrl(outputItem.getOutputPath())));
            items.add(new PostCompileProcessingItem(vf, packageName, outputItem.getSourceFile()));
          }
        }
      });
    }
    catch(Exception ex) {
      throw new RuntimeException(ex);
    }
    return items;
  }
  
  protected List/*<ModuleTaggedProcessingItem>*/ getModuleTaggedProcessingItems(final CompileContext compileContext) {
    final List items = new ArrayList(); // List of items to return
    try {
      ActionRunner.runInsideReadAction(new ActionRunner.InterruptibleRunnable() {
        public void run() {
          // Map paths to modules, to know which module each file belongs to
          final Map pathModules = new HashMap();
          final CompileScope compileScope = compileContext.getCompileScope();
          Module[] modules = compileScope.getAffectedModules();
          for (int i = 0; i < modules.length; i++) {
            Module module = modules[i];
            VirtualFile output = compileContext.getModuleOutputDirectory(module);
            if(output != null) {
              pathModules.put(output, module);
              LOG.debug(output.getPath() + " points to " + module.getName());
            }
            VirtualFile testOutput = compileContext.getModuleOutputDirectoryForTests(module);
            if(testOutput != null) {
              pathModules.put(testOutput, module);
              LOG.debug(testOutput.getPath() + " points to " + module.getName());
            }
          }

          for(Iterator<RecentlyCompiledSink.CompiledItem> iter = getRecentlyCompiled(compileContext).iterator(); iter.hasNext(); ) {
            RecentlyCompiledSink.CompiledItem outputItem = iter.next();
            if (outputItem.getOutputPath() == null) { // Needed for package-info.java
              compileContext.addMessage(CompilerMessageCategory.WARNING, "AspectJ: Skipping " + outputItem.getSourceFile(), null, -1 ,-1);
              continue;
            }
            
            String packageName = PluginUtils.resolvePackageName(outputItem.getOutputRoot(), outputItem.getOutputPath());
            String classFileName = outputItem.getOutputPath();
            // LOG.info("Adding " + classFileName + " to processing items. Package: " + packageName);
            // final Module module = compileContext.getModuleByFile(vf);
            // Find module for compiled file
            Module module = null;
            for(Iterator moduleIter = pathModules.entrySet().iterator(); moduleIter.hasNext(); ) {
              Map.Entry moduleEntry = (Map.Entry) moduleIter.next();
              final VirtualFile modulePath = (VirtualFile)moduleEntry.getKey();
              if(modulePath != null && classFileName.startsWith(modulePath.getPath()))
                module = (Module) moduleEntry.getValue();
            }
            if(module == null)
              LOG.info("No module found for: " + classFileName);
            // (Requires read action)
            VirtualFile vf = VfsUtil.findFileByURL(
                VfsUtil.convertToURL(VfsUtil.pathToUrl(classFileName)));
            items.add(new ModuleTaggedProcessingItem(vf, packageName, outputItem.getSourceFile(), module));
            /*
            try {
              VirtualFile vf = VfsUtil.findFileByURL(new File(classFileName).toURL());
              items.add(new ModuleTaggedProcessingItem(vf, packageName, module));
            }
            catch(MalformedURLException muex) {
              throw new RuntimeException(muex);
            }
            */
          }
        }
      });
    }
    catch(Exception ex) {
      throw new RuntimeException(ex);
    }
    return items;
  }

  // IDEA v4
  public final boolean validateConfiguration() {
    return true;
  }

  // IDEA v5-6
  public final boolean validateConfiguration(CompileScope compileScope) {
    return true;
  }

  public ValidityState createValidityState(DataInputStream is) throws IOException {
    // Not quite sure what this does...
    return new TimestampValidityState(is.readLong()); // From RetroWeaver plugin
    // return new TimestampValidityState(System.currentTimeMillis()); // From Backport75 plugin
  }

  // IDEA v 8
  public ValidityState createValidityState(DataInput di) throws IOException {
    // Not quite sure what this does...
    return new TimestampValidityState(di.readLong()); // From RetroWeaver plugin
    // return new TimestampValidityState(System.currentTimeMillis()); // From Backport75 plugin
  }

  /**
   * Sort the files per module.
   */
  protected static Map/*<Module,List<ModuleTaggedProcessingItem>>*/ sortItemsPerModule(final ProcessingItem[] processingItems) {
    final Map filesPerModule = new LinkedHashMap();
    for (int i = 0; i < processingItems.length; i++) {
      // if (processingItem.getFile().getPath().endsWith(".class")) { // Should not be neede b.o. WrappingCompiler
      final ModuleTaggedProcessingItem processingItem = (ModuleTaggedProcessingItem) processingItems[i];
      Module module = processingItem.getModule();
      List fileInThisModule = (List) filesPerModule.get(module);
      if(fileInThisModule == null) {
        fileInThisModule = new ArrayList();
        filesPerModule.put(module, fileInThisModule);
      }
      fileInThisModule.add(processingItem);
    }
    return filesPerModule;
  }
}