/**
 * [BSD Open Source License]
 *
 * Copyright (c) 2006, Mattias Jiderhamn, Expert Systems
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or othermaterials provided with the distribution.
 *
 * Neither the name of Expert Systems nor the names of its contributors may
 * be usedto endorse or promote products derived from this software without
 * specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package se.expertsystem.intellij;

import java.util.Collections;
import java.util.HashSet;

import com.intellij.openapi.compiler.*;
import com.intellij.openapi.compiler.Compiler;
import com.intellij.openapi.diagnostic.Logger;
import com.intellij.openapi.fileTypes.FileType;
import com.intellij.openapi.fileTypes.StdFileTypes;
import com.intellij.openapi.project.Project;
import com.intellij.openapi.util.InvalidDataException;
import com.intellij.openapi.util.JDOMExternalizer;
import com.intellij.openapi.util.WriteExternalException;
import com.intellij.openapi.util.JDOMExternalizable;

import org.jdom.Element;

/**
 * Base class for project components that add AOP weaving to IntelliJ compilation.
 * @author Mattias Jiderhamn
 */
public abstract class PostProcessProjectComponent implements ToggledProjectComponent, JDOMExternalizable {
  
  private static final Logger LOG = Logger.getInstance(PostProcessProjectComponent.class.getName());

  /** The project that this component is bound to */
  protected Project project;

  /** Is this component activated? */
  protected boolean isActivated = false;

  /** The instrumenting compiler that does the weaving */
  protected PostProcessCompiler postCompiler;

  public Project getProject() {
    return project;
  }

  /** Make sure regular Java compiler is wrapped */
  private void wrapCompiler() {
    com.intellij.openapi.compiler.Compiler[] compilers = getCompilerManager().getCompilers(TranslatingCompiler.class); // Compiler.class
    for(int i = 0; i < compilers.length; i++) {
      Compiler compiler = compilers[i];
      if(compiler.getClass().getName().equals(WrappingCompiler.class.getName())) {
        LOG.debug("Wrapping java compiler added by other plugin than " + getComponentName());
        break; // Already wrapped by other plugin
      }
      else if("Java Compiler".equals(compiler.getDescription())) {
        // Wrap regular Java compiler
        WrappingCompiler wrappingCompiler = new WrappingCompiler((TranslatingCompiler) compiler);
        getCompilerManager().removeCompiler(wrappingCompiler.getRealCompiler()); // Remove real compiler
        getCompilerManager().addTranslatingCompiler(wrappingCompiler, // Add wrapping compiler 
            new HashSet<FileType>(Collections.singleton(StdFileTypes.JAVA)),
            Collections.singleton(StdFileTypes.CLASS)
        );
        LOG.debug("Java compiler wrapped by " + getComponentName());
      }
    }
  }

  /** Get compiler manager for current project */
  private CompilerManager getCompilerManager() {
    return CompilerManager.getInstance(project);
  }

  /** Change state of component */
  public void toggle() {
    changeTo(!isActivated);
  }

  /** Change state of component to provided state */
  public void changeTo(boolean activated) {
    if (activated != isActivated) {
      if (activated) {
        wrapCompiler(); // Wrap compiler

        getCompilerManager().addCompiler(postCompiler); // Add weaving compiler
        LOG.debug("Added compiler " + postCompiler);
      } else {
        // (Never unwrap, since other plugins may use the wrapped compiler)

        getCompilerManager().removeCompiler(postCompiler); // Remove weaving compiler
        LOG.debug("Removed compiler " + postCompiler);
      }
      isActivated = activated;
    }
  }
  
  protected void configLoaded(boolean activated) {
    changeTo(activated); // No special treatment
  }

  public boolean isActivated() {
    return isActivated;
  }

  /**
   * Read on/off state from IWS file
   */
  public void readExternal(Element element) throws InvalidDataException {
    boolean config = JDOMExternalizer.readBoolean(element, "isActivated");
    configLoaded(config);
  }

  /**
   * Persists on/off state in IWS file
   */
  public void writeExternal(Element element) throws WriteExternalException {
    JDOMExternalizer.write(element, "isActivated", isActivated);
  }
}