/**
 * [BSD Open Source License]
 *
 * Copyright (c) 2006, Mattias Jiderhamn, Expert Systems
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer. 
 * 
 * Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or othermaterials provided with the distribution. 
 * 
 * Neither the name of Expert Systems nor the names of its contributors may
 * be usedto endorse or promote products derived from this software without
 * specific prior written permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */ 
package se.expertsystem.intellij;

import com.intellij.openapi.compiler.TranslatingCompiler;
import com.intellij.openapi.compiler.CompileContext;
import com.intellij.openapi.compiler.CompileScope;
import com.intellij.openapi.vfs.VirtualFile;
import com.intellij.openapi.diagnostic.Logger;
import com.intellij.openapi.module.Module;
import com.intellij.util.Chunk;

import java.lang.reflect.Method;
import java.util.List;

/**
 * This compiler wraps the regular Java Compiler to remember the compiled classes.
 * @author Mattias Jiderhamn, mattias.jiderhamn@expertsystems.se
 */
public class WrappingCompiler implements TranslatingCompiler {
  
  private static final Logger LOG = Logger.getInstance(WrappingCompiler.class.getName());

  /** The actual Java compiler */
  private final TranslatingCompiler realCompiler;

  public TranslatingCompiler getRealCompiler() {
    return realCompiler;
  }

  public WrappingCompiler(TranslatingCompiler realCompiler) {
    this.realCompiler = realCompiler;
  }

  /**
   * Checks if the compiler can compile the specified file.
   * Question will be redirected to wrapped Java compiler
   */
  public boolean isCompilableFile(VirtualFile virtualFile, CompileContext compileContext) {
    return realCompiler.isCompilableFile(virtualFile, compileContext);
  }

  /**
   * Compiles the specified files.
   * Actual compilation will be performed by wrapped Java compiler. After compilation, all the recompiled files
   * will be reported to the AOP weaving compiler.
   */
  public void compile(CompileContext compileContext, Chunk<Module> moduleChunk, VirtualFile[] virtualFiles, OutputSink sink) {
	  RecentlyCompiledSink wrappedSink = new RecentlyCompiledSink(sink);

    // Perform actual compilation
    realCompiler.compile(compileContext, moduleChunk, virtualFiles, wrappedSink);

	  List<RecentlyCompiledSink.CompiledItem> recentlyCompiled = PostProcessCompiler.getRecentlyCompiled(compileContext);
	  recentlyCompiled.addAll(wrappedSink.getCompiledItems());
  }

  public String getDescription() {
    return "Wrapped " + realCompiler.getDescription();
  }

  // IDEA 5-6
  public boolean validateConfiguration(CompileScope compileScope) {
    // return realCompiler.validateConfiguration(compileScope);
    try {
      Method m = realCompiler.getClass().getMethod("validateConfiguration", new Class[] {CompileScope.class});
      return ((Boolean)m.invoke(realCompiler, new Object[] {compileScope})).booleanValue();
    }
    catch(Exception ex) {
      LOG.error(ex);
      return false;
    }
  }

  // IDEA 4.5
  public boolean validateConfiguration() {
    // return realCompiler.validateConfiguration();
    try {
      Method m = realCompiler.getClass().getMethod("validateConfiguration", null);
      return ((Boolean)m.invoke(realCompiler, null)).booleanValue();
    }
    catch(Exception ex) {
      LOG.error(ex);
      return false;
    }
  }
}